import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import com.partnersoft.system.SystemServices;
import com.partnersoft.data.*;
import com.partnersoft.io.vfs.*;
import com.partnersoft.system.RuntimeLib;
import com.partnersoft.net.NetworkService;
import com.partnersoft.modules.upgradeservice.lib.SettingsLib;
import com.partnersoft.sql.account.*;


/**
 * See #1204.  Will run mysql_upgrade if it appears the tables have never been upgraded.
 */


/**
 * Create a new Vfs rooted above the MasterInstall and CentralHub directories.
 * @return a Vfs above both MI and CH 
 */
private FileVfs getPartnerHubVfs() {
	//this guy is gonna be running out of the master install, we need to list directories in CentralHub/data/MySQL
	//get the vfs that we are running in
	Vfs currentVfs = SystemServices.vfs();

	//get the parent of master install
	Path hubPath = ((Path)currentVfs.getAbsolutePath()).getParent();
	String hubPathString = hubPath.toFileString();
	//create a new vfs above both master install and central hub
	FileVfs hubVfs = new FileVfs(hubPathString);
	return hubVfs;
}

/**
 * A copy of RuntimeLib.execAndWait(String[] commandArray, File currentDir),
 * with the key distinction that it does not log it's arguments.
 * This prevents passwords from being written into the log.
 */
private int execAndWaitQuietly(String[] commandArray, File currentDir)
		throws IOException {
	List<String> commandList = Arrays.asList(commandArray);
	// ssshhhh....
	// log.info("Running external commands: " + commandList);
	if (currentDir != null)
		log.info("Current directory is: " + currentDir);

	try {
		Process proc = Runtime.getRuntime().exec(commandArray, null,
				currentDir);
		RuntimeLib.logOutput(proc);
		// ssshhhh....
		// log.info("Waiting for external command " + commandList + " to complete.");
		int exitVal = proc.waitFor();
		log.info("Exec exit value: " + exitVal);
		return exitVal;
	} catch (Exception oopsie) {
		log.fatal("Error running external command: " + commandList[0], oopsie);
		return -1;
	}
}

private boolean isMissingTables(VfsDirectory dbTablesDir) {

	ArrayList<String> foundTables = new ArrayList();

	// just list the MYI files, store the base names
	for(tableFile in dbTablesDir.listFilesWithExtension("myi"))
		foundTables.add(tableFile.getBaseName());

	// check if the table roles_mapping is missing
	if(!foundTables.contains("roles_mapping")) {
		log.warn("Missing table \'roles_mapping\'.");
		return true;
	}

	return false;
}

/**
 * Tests to see if the MySQL TCP port 3306 is open and available - generally a good indication that
 * MySQL is not running.
 * @return true if port is available, false otherwise.
 */
private boolean isPortAvailable() {
	try {
		log.info("Checking to see if MySQL port " + service.getPortNumber() + " is occupied.");
		Socket sock = new Socket("127.0.0.1", service.getPortNumber());
		sock.close();
		log.info("Port seems to be there, assuming MySQL is on.");
		return false;
	}
	catch (Exception oopsie) {
		log.info("Caught an " + oopsie + "; assuming MySQL is not on.");
		return true;
	}
}

private boolean isDaemonRunning() {
	return !isPortAvailable();
}

private boolean napForVictory(boolean start) {
	
	String action = "stop";
	if(start)
		action = "start";
	

	for (int i = 0; i < 10; i++) {

		if(isDaemonRunning()) {
			// if the daemon is running and we wanted it to start, then success
			if(start)
				return true;
		}else {
			// if the daemon is NOT running and we wanted it to stop, then success
			if(!start)
				return true;
		}
		try {
			log.info("Napping to give MySQL a chance to " + action + ".");
			Thread.sleep(1000);
		}
		catch (InterruptedException oopsie) {
			log.warn("Zzzzz...snrk!?!");
		}
	}
	
	return false;
}

private boolean startMariaDb() {
	StringBuilder args = new StringBuilder();
	args.append(" --defaults-file=\"");
	args.append(configFile.getAbsolutePath().toFileString());
	args.append("\" --standalone --console");

	log.info("Starting mysqld.exe with args:\n" + args);


	mysqld = mySqlBinaryDir.fileNamed("mysqld.exe");
	RuntimeLib.execNoWait(mysqld.toFile().getCanonicalPath() + args.toString(), mySqlBinaryDir.toFile());

	return napForVictory(true);
}

private boolean stopMariaDb() {
	mysqladmin = mySqlBinaryDir.fileNamed("mysqladmin.exe");
	
	Account dbShutdown = DatabaseAccountFactory.createDbShutdownAccount();
	
	StringBuilder args = new StringBuilder();
	args.append(" --host=localhost");
	args.append(" --port=");
	args.append(service.getPortNumber());
	args.append(" --user=");
	args.append(dbShutdown.getAccountName());
	args.append(" --password=");
	args.append(dbShutdown.getPassword());
	args.append(" shutdown");
	
	log.info("Stopping mysqld daemon.");
	
	RuntimeLib.execNoWait(mysqladmin.toFile().getCanonicalPath() + args.toString(), mySqlBinaryDir.toFile());
	return napForVictory(false);
}

private void upgradeTables() {
	
	// build path to upgrade binary
	upgrader = mySqlBinaryDir.fileNamed("mysql_upgrade.exe");
	
	// get admin user to run upgrade
	Account partnerAdmin = DatabaseAccountFactory.createPartnerAdminAccount();
	
	String []commandArray = new String[3];
	commandArray[0] = upgrader.toFile().getAbsolutePath();
	commandArray[1] = "-u" + partnerAdmin.getAccountName();
	commandArray[2] = "-p" + partnerAdmin.getPassword();
	
	int result = -1;
	try {
		result = execAndWaitQuietly(commandArray, mySqlBinaryDir.toFile());
	}catch(Exception e) {
		log.error("Exception while trying to upgrade tables!\n" + commandArray[0], e);
	}
	
	if(result == 0) {
		log.info("Successfully upgraded MySQL tables!");
	} else if(result == 1) {
		log.info("Already upgraded!");
	} else {
		log.error("Failed to upgrade MySQL tables, manual intervention is required.");
		criticals.add("Failed to upgrade MySQL tables, manual intervention is required.");
	}
}
//////////		Start Main 			/////////////////

FileVfs hubVfs = getPartnerHubVfs();

// following are globals that are avaiable to methods
centralHubDir = hubVfs.directoryFor("CentralHub");
//get data dir
centralHubMySqlDataDir = centralHubDir.directoryFor("data/MySQL");

// see if we need to run mysql_upgrade.
// do this by checking for tables that should be present after AMP upgrade
dbTablesDir = centralHubMySqlDataDir.directoryFor("mysql");
if( !isMissingTables(dbTablesDir) ) {
	log.info("No need to run upgrade_tables.");
	return;
}

log.warn("Will attempt to run mysql_upgrade");

// initialize some other stuff needed to run mysql

service = NetworkService.HUB_MYSQL;

// get instance of MasterInstall MySQL module
mySqlModule = SystemServices.moduleManager().moduleNamed("MySQL");
// get the bin directory
mySqlBinaryDir = mySqlModule.getDirectory().directoryFor("os/windows/applications/MySQL/bin/");

// previous script should have created a one-time config file to use
configFile = mySqlModule.getTempDirectory().fileFor("partner.cnf");
	
// check if MariaDb is running, start if necessary
if(!isDaemonRunning()) {
	// if MariaDb failed to start, shit the pants
	if(!startMariaDb()) {
		criticals.add("Unable to launch the mysql daemon!  Please check configuration and other logs.");
		return;
	}
}

upgradeTables();

if(!stopMariaDb()) {
	log.error("Failed to stop MariaDb!");
	criticals.add("Failed to stop MariaDb, this is also bad.");
}
